#!/usr/bin/python3

import numpy as np
from matplotlib import pyplot as plt
from .Layer import Layer
from .Dataset import Dataset
import json
import time
from .LossFunctions import LossFunctions
import typing
from .Types import T_Feature_Array, T_Target_Array, T_Output_Array, T_Data_Sample, T_Dataset
from tqdm import tqdm
from . import Utils

np.set_printoptions(precision=20)


class NeuralNetwork:
    # Main NeuralNetwork class
    def __init__(self, I, O, cost="mse"):
        """
        Creates a Feed Forward Neural Network.

        Parameters
        ----------
        I : int
            Number of inputs to the network

        O : int
            Number of outputs from the network

        [cost]: string
            The cost/loss function used by the neural network.
            Default value is 'mse' which stands for Mean Squared Error.

            Available options:
                mse => Mean Squared Error
                ce => Cross Entropy

        Returns
        -------
        Doesn't return anything
        """

        # print("Construct")
        self.Network: typing.List[Layer] = list()
        self.I = I
        self.O = O
        self.cost = cost
        self.loss_computer = LossFunctions(cost)
        self.total_layers = 0
        self.learning_rate = 0.01
        self.is_loaded_model = False
        self.model_compiled = False
        self.prediction_evaulator = Utils.judge_prediction

    def set_learning_rate(self, lr):
        """
        Modifies the learning rate of the network.

        Parameters
        ----------
        lr : float
            New learning rate

        Returns
        -------
        Doesn't return anything
        """
        self.learning_rate = lr

    def add_layer(self, num_nodes, activation_function="sigmoid"):
        """
        Adds a layer to the network.

        Parameters
        ----------
        num_nodes : int
            Number of nodes in the hidden layer

        [activation_function] :str
            It is an optional parameter.
            Specifies the activation function of the layer.
            Default value is sigmoid.

            Available options:
                sigmoid,
                tanh,
                linear,
                identity,
                softmax

        Returns
        -------
        Doesn't return anything
        """
        # A layer can be thought of as a matrix
        # No. of row = no. of nodes
        # No. of columns = No. of weights = No. of inputs + 1 (bias)
        if self.total_layers == 0:
            inputs = self.I
        else:
            last_layer = self.Network[-1]
            inputs = last_layer.num_nodes
        layer = Layer(num_nodes, inputs, activation_function, self.cost)
        self.Network.append(layer)
        self.total_layers += 1

    def compile(self, activation_function="sigmoid"):
        """
        Basically, it just adds the output layer to the network.

        Parameters
        ----------
        [activation_function] :str
            It is an optional parameter.
            Specifies the activation function of the layer.
            Default value is sigmoid.

        Returns
        -------
        Doesn't return anything
        """
        if self.model_compiled:
            print("[!!] Model is already compiled!")
            print("[!!] You cannot add layers anymore")
            return
        self.is_loaded_model = False
        self.model_compiled = True
        # Adding output layer
        self.add_layer(self.O, activation_function=activation_function)

        # for layer in self.Network :
        #     print(np.mean(layer.weights), np.std(layer.weights))
        #     print(np.mean(layer.biases), np.std(layer.biases))
        #     print()

    def feedforward(self, input_array: T_Feature_Array):
        """
        Feeds the given input throughout the network

        Parameters
        ----------
        input_array : T_Feature_Array
            Input to be fed to the network.
            It is columnar vector of size Inputs x 1

        Returns
        -------
        all_outputs : T_Output_Array
            An array of all the outputs produced by each layer.
        """
        all_outputs: typing.List[T_Output_Array] = list()
        _i = 1
        for layer in self.Network:
            # print("Feeding ", input_array.T, "to , layer", i)
            outputs = layer.feed(input_array)
            all_outputs.append(outputs.T)
            input_array = outputs
            # print("All outputs: ", all_outputs)
            # print()
            # i += 1
        return all_outputs

    def backpropagate(self, target: T_Target_Array):
        """
        Backpropagate the error throughout the network
        This function is called inside the model only.

        Parameters
        ----------
        target : np.array()
            It is the ground truth value corresponding to the input.
            It is columnar vector of size Outputs x 1

        Returns
        -------
        Error : float
            # Returns the Mean Squared Error of the particular output
            Returns the error using the specified loss function.
        """
        for i in range(self.total_layers - 1, -1, -1):
            layer = self.Network[i]
            if i == self.total_layers - 1:
                # print("Output layer: ", layer.output_array, "Target: ", target)
                output_errors = self.loss_computer.get_loss(
                    layer.output_array, target)

                # Evaluate this prediction to compute accuracy
                is_correct_output: bool = self.prediction_evaulator(
                    layer.output_array, target)

                # print("Error: ", output_errors)
                layer.calculate_gradients(target, "output")
            else:
                next_layer = self.Network[i + 1]
                layer.calculate_gradients(
                    next_layer.weights, "hidden", next_layer.deltas
                )
        return output_errors, is_correct_output

    def update_weights(self, input_array: T_Feature_Array):
        """
        Update the weights of the network.
        This function is called inside the model only.

        Parameters
        ----------
        input_array : np.array()
            It is the input fed to the network
            It is columnar vector of size Inputs x 1

        Returns
        -------
        Doesn't return anything
        """
        for i in range(self.total_layers - 1, -1, -1):
            layer = self.Network[i]
            if i == 0:
                # if it is the first layer => inputs = input_array
                layer.update_weights(input_array, self.learning_rate)
            else:
                # not the first most => inputs = previous layer's output
                inputs = self.Network[i - 1].output_array
                layer.update_weights(inputs, self.learning_rate)

    def Train(self, dataset: T_Dataset, size, epochs=100, logging=False, epoch_logging=True, prediction_evaulator=None):
        """
        Trains the neural network using the given dataset.

        Parameters
        ----------
        dataset : T_Dataset

        size : int
            Size of the dataset

        [epochs] : int
            An optional parameter.
            Number of epochs to train the network. Default value is 5000

        [logging] : bool
            An optional parameter.
            If its true, all outputs from the network will be logged out onto STDOUT for each epoch.

        [epoch_logging] : bool
            An optional parameter.
            If it is true, Error in each epoch will be logged to STDOUT.

        [prediction_evaulator]: (prediction: T_Output_Array, target: T_Output_Array) -> bool
            An optional parameter.
            Used to evaluate the fed forward output with the actual target.
            Default value is 'Utils.judge_prediction' function.

        Returns
        -------
        Doesn't return anything.
        """

        if prediction_evaulator is not None:
            self.prediction_evaulator = prediction_evaulator

        if not self.model_compiled:
            print("[-] Network is not complete.!")
            print("[!!] Please compile the network by adding an output layer")
            return

        self.all_errors = list()
        self.epochs = epochs

        # Show progress bar if epoch_logging is False
        range_object = range(epochs)
        if not epoch_logging:
            range_object = tqdm(range(epochs), desc="Training...")

        for epoch in range_object:
            self.loss = 0
            # One Epoch
            if epoch == 0:
                start = time.time()
            elif epoch == 1:
                end = time.time()
                estimated_time = Utils.get_time_required(start, end, epochs)
                print(estimated_time)

            correct = 0
            for i in range(size):
                data_sample = dataset[i]
                # input_array = data_sample[0]
                # target_array = data_sample[1]
                input_array = data_sample[0]
                target_array = data_sample[1]

                all_outputs = self.feedforward(input_array)
                output_error, is_correct_output = self.backpropagate(
                    target_array)
                self.update_weights(input_array)

                if is_correct_output:
                    correct += 1
                self.loss += output_error
                if logging:
                    print(
                        input_array.transpose(),
                        "\x1b[35m",
                        all_outputs,
                        "\x1b[0m",
                        target_array.transpose(),
                        "\x1b[31m",
                        output_error,
                        "\x1b[0m",
                    )

            self.loss /= size
            self.all_errors.append(self.loss)
            self.accuracy = (correct*100)/size

            if epoch_logging:
                print(
                    f"Epoch: {epoch + 1} ==> ({self.cost}) Error: {self.loss}, (%) Accuracy: {self.accuracy}")

            if logging:
                print()

    def predict(self, input_array: T_Feature_Array):
        """
        Predicts the output using a given input.

        Parameters
        ----------
        input_array : np.array()
            It is columnar vector of size Inputs x 1
            It is the input fed to the network

        Returns
        -------
        prediction : np.array()
            Predicted value produced by the network.
        """
        return self.feedforward(input_array)[-1]

    def epoch_vs_error(self):
        """
        Plot error vs epoch graph

        Parameters
        ----------
        Doesn't accept any parameters

        Returns
        -------
        Doesn't return anything
        """
        if self.is_loaded_model:
            print("[!!] You cannot look at epoch vs error graph in a loaded model")
            print("[!!] You can only look at that while training.!")
            print("[!!] Make some modifications to the network to own the model")
            return
        all_epochs = [i + 1 for i in range(self.epochs)]
        plt.xlabel("Epoch")
        plt.ylabel("Error")
        plt.title("Epoch vs Error")
        plt.plot(all_epochs, self.all_errors)
        plt.show()

    def evaluate(self):
        """
        Print the basic information about the network.
        Like accuracy, error ..etc.

        Parameters
        ----------
        Doesn't accept any parameters

        Returns
        -------
        Doesn't return anything
        """
        print("\t=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-")
        print("\tModel is trained for ", self.epochs, "epochs")
        print("\tModel Accuracy: ", self.accuracy, "%")
        if self.accuracy < 70:
            print("\t\tModel Doesn't seem to have fit the data correctly")
            print("\t\tTry increasing the epochs")
        print("\t=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-")

    def display(self):
        """
        Print the information of each layer of the network.
        It can be used to debug the network!

        Parameters
        ----------
        Doesn't accept any parameters

        Returns
        -------
        Doesn't return anything
        """
        for i in range(self.total_layers):
            print("Layer: ", i + 1)
            self.Network[i].display()

    def export_model(self, filename):
        """
        Export the model to a json file

        Parameters
        ----------
        filename: str
            File name to export model

        Returns
        -------
        Doesn't return anything
        """
        try:
            fhand = open(filename, "w")
        except Exception as e:
            print("[!!] Unable to open file ", filename, ": ", e)
            print("[!!] Couldn't export model")
            return

        model_info = dict()
        model_info["inputs"] = self.I
        model_info["outputs"] = self.O
        model_info["learning_rate"] = self.learning_rate
        model_info["model_compiled"] = self.model_compiled
        model_info["layers"] = list()
        for layer in self.Network:
            layer_object = dict()
            layer_object["neurons"] = layer.num_nodes
            layer_object["inputs"] = layer.inputs
            layer_object["weights"] = layer.weights.tolist()
            layer_object["biases"] = layer.biases.tolist()
            layer_object["activation_function"] = layer.activation_function
            layer_object["loss_function"] = layer.loss_function
            model_info["layers"].append(layer_object)
        model_info["accuracy"] = self.accuracy
        model_info["loss"] = self.loss
        model_info["epochs"] = self.epochs

        json_format_string = json.dumps(model_info)
        fhand.write(json_format_string)
        fhand.close()
        print("[*] Model exported successfully to", filename)

    @staticmethod
    def load_model(filename):
        """
        Load model from an eported (json) model

        Parameters
        ----------
        filename : str
            Exported model (json) file

        Returns
        -------
        brain : NeuralNetwork
            NeuralNetwork object
        """
        try:
            fhand = open(filename, "r")
        except Exception as e:
            print("[!!] Unable to open file ", filename, ": ", e)
            print("[!!] Couldn't load model")
            return

        model_info = json.load(fhand)
        # print(model_info)

        inputs = model_info["inputs"]
        outputs = model_info["outputs"]

        brain = NeuralNetwork(inputs, outputs)
        brain.total_layers = 0
        brain.is_loaded_model = True

        for layer_object in model_info["layers"]:
            num_nodes = layer_object["neurons"]
            weights = layer_object["weights"]
            inputs = layer_object["inputs"]
            biases = layer_object["biases"]
            activation_fn = layer_object["activation_function"]
            loss_function = layer_object["loss_function"]

            layer = Layer(
                num_nodes, inputs, activation_function=activation_fn, loss_function=loss_function)
            layer.weights = np.array(weights)
            layer.biases = np.array(biases)
            brain.Network.append(layer)
            brain.total_layers += 1

        brain.accuracy = model_info["accuracy"]
        brain.loss = model_info["loss"]
        brain.epochs = model_info["epochs"]
        brain.learning_rate = model_info["learning_rate"]
        brain.model_compiled = model_info["model_compiled"]
        print("[*] (", filename, ") Model Loaded successfully", sep="")

        return brain



import numpy as np
from .ActivationFunction import ActivationFunction
from .LossFunctions import LossFunctions
from .Types import T_Feature_Array, T_Output_Array

# Layer class


class Layer:
    def __init__(self, num_nodes, inputs, activation_function, loss_function):
        """
            Layer constructor

            Parameters
            ----------
            num_nodes : int
                No. of nodes in the layer

            inputs : int
                No. of inputs to the layer

            activation_function

            Returns
            -------
            None
        """
        self.inputs = inputs
        self.num_nodes = num_nodes
        self.weights = np.random.randn(
            num_nodes, inputs) * np.sqrt(2 / num_nodes)
        self.biases = np.random.randn(num_nodes, 1)
        self.output_array = np.random.randn(num_nodes, 1)
        self.deltas = np.zeros((num_nodes, 1))
        self.activation_function = activation_function
        self.activator = ActivationFunction(self.activation_function)
        self.loss_function = loss_function
        self.loss_computer = LossFunctions(self.loss_function)

    def feed(self, input_array: T_Feature_Array) -> T_Output_Array:
        """
            Feeds the given input array to a particular layer.

            Parameters
            ----------
            input_array: T_Feature_Array
                Input array to be fed to the layer

            Returns
            -------
            output_array: T_Output_Array
        """
        # print("Weights", self.weights.shape)
        # print("Inputs ", input_array.shape)
        self.input_array = input_array
        dot_product = np.dot(self.weights, input_array)
        dot_product += self.biases
        # print("Output: ", dot_product.shape)
        output_array = self.activate(dot_product)
        self.output_array = output_array  # Store the output in the layer.
        # print(self.input_array.shape, self.output_array.shape)
        return output_array

    def activate(self, x):
        """
            Passes the output array to an activation function.

            Parameters
            ----------
            x
                Output array from a layer

            Returns
            -------
            Activated output
        """
        return self.activator.activate(x)

    def calculate_gradients(self, target_or_weights, layer_type, next_layer_deltas=None):
        """
            Calculates the gradients for each weight and bias

            Parameters
            ----------
            target_or_weights
                This is either targers array of weights matrix.
                Specifically, it'll be the targets array while computing the gradients for the output layer
                and weights matrix of the next layer.

            layer_type
                This will either be "hidden" or "output"

            [next_layer_deltas]
                This is (not exactly) an optional parameter.
                This will be passed only while computing the gradients of a hidden layer.

            Returns
            -------
                Doesn't return anything. But stores the gradients as a class attribute.
        """

        activation_gradient = self.activator.activate(
            self.output_array, derivative=True)
        if layer_type == "output":
            # print("Output Layer")
            # target_or_weights => target values
            loss_gradient = self.loss_computer.get_gradient(
                self.output_array, target_or_weights)
            self.deltas = loss_gradient * activation_gradient

            # delta = (target - output) * activation_derivative(output)
        else:
            # print("Hidden Layer")
            # target_or_weights = Next layer's weights
            hidden_errors = np.dot(target_or_weights.transpose(
            ), next_layer_deltas)  # Errors in the hidden layer
            self.deltas = hidden_errors * activation_gradient
            # deltas += (next_layers weights * next_layers deltas)

    def update_weights(self, inputs, learning_rate):
        """
            Tweak the weights of the layer.

            Parameters
            ----------
            inputs: T_Feature_Array
                Input to this network

            learning_rate: float
                Learning rate of the entire network.

            Returns
            -------
            Doesn't return anything.
        """

        change_in_weights = np.dot(self.deltas, inputs.T) * learning_rate
        self.weights = np.subtract(self.weights, change_in_weights)
        self.biases = np.subtract(self.biases, self.deltas * learning_rate)

    def display(self):
        """
            Display the metadata of the layer.
        """

        print("\tInputs: ", self.inputs)
        print("\tWeights: ", self.weights)
        print("\tBiases: ", self.biases.T)
        print("\tDeltas: ", self.deltas.T)
        print("\toutput_array: ", self.output_array.T)
        print("\tActivation: ", self.activation_function)



import numpy as np


class LossFunctions:
    def __init__(self, loss_function):
        self.loss_function = loss_function
        self.function = self._mse

        if loss_function == 'mse':
            self.function = self._mse
        elif loss_function == 'ce':
            self.function = self._cross_entropy

    def get_loss(self, outputs, targets):
        # print("Outputs:", outputs)
        # print("Targets:", targets)
        loss = self.function(outputs, targets)
        # print("loss:", loss)
        # print()
        return loss

    def get_gradient(self, outputs, targets):
        # print("Outputs:", outputs)
        # print("Targets:", targets)
        gradient = self.function(outputs, targets, derivative=True)
        # print("gradient:", gradient)
        # print()
        return gradient

    def _mse(self, outputs, targets, derivative=False):
        # print("hihi")
        error = np.subtract(targets, outputs)
        if derivative:
            return -1*error
        return np.mean(error*error)

    def _cross_entropy(self, outputs, targets, derivative=False):
        error = np.subtract(targets, outputs)
        if derivative:
            return -1*error
        # print("cross_entropy:", targets*np.log(outputs))
        # print("cross_entropy:", np.sum(targets*np.log(outputs)))
        return -1*np.sum(targets*np.log(outputs))



import random
from math import floor
import numpy as np
from .Types import T_Output_Array


def shuffle_array(array: list):
    array_copy = array.copy()
    random.shuffle(array_copy)

    return array_copy


def split_arr(array: list, ratio: float):
    n = len(array)

    m = floor(n * ratio)

    first_part: list = array[0: m]
    second_part: list = array[m: n]

    return [first_part, second_part]


def one_hot_encode(num, size):
    vector = np.array([0]*size)
    vector[num-1] = 1

    return vector.reshape(size, 1)


def get_time_required(start, end, epochs):
    seconds = (end - start) * epochs

    minutes = seconds // 60
    seconds = seconds % 60

    hours = minutes // 60
    minutes = minutes % 60
    return f"Estimated Training Time: {hours}hrs::{minutes}min::{round(seconds, 4)}sec"


def judge_prediction(prediction: T_Output_Array, target: T_Output_Array) -> bool:
    is_correct_output: bool = np.argmax(prediction) == np.argmax(target)
    return is_correct_output



#!/usr/bin/python3

from setuptools import setup, find_packages


with open("README.md", 'r') as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as fh:
    install_requires = fh.read().splitlines()


print(find_packages())

setup(
    name='nicenet',
    version='0.0.7',
    author="Subhash Sarangi",
    author_email="subhashsarangi123@gmail.com",
    url="https://github.com/Subhash3/Neural_Net_Using_NumPy/",
    packages=["nicenet"],
    description='Feed Forward Neural Networks',
    long_description=long_description,
    long_description_content_type="text/markdown",
    # py_modules=["nicenet"],
    # package_dir={'': 'nicenet'},
    install_requires=install_requires,
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        #"License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: OS Independent",
    ],
)



import numpy as np
from typing import TypeVar, Generic

T = TypeVar('T')

# Class for activation functions


class ActivationFunction(Generic[T]):
    def __init__(self, function_name):
        if function_name == 'sigmoid':
            self.activation_function = self._sigmoid
        elif function_name == 'tanh':
            self.activation_function = self._tanh
        elif function_name == 'relu':
            self.activation_function = self._relu
        elif function_name == 'softmax':
            self.activation_function = self._softmax
        elif function_name == "identity":
            self.activation_function = self._identity

    def activate(self, x: T, derivative=False) -> T:
        if derivative:
            return self.activation_function(x, derivative=True)
        return self.activation_function(x)

    def _sigmoid(self, x: T, derivative=False) -> T:
        if derivative:
            # y = self._sigmoid(x)
            y = x
            return y * (1 - y)
        return 1/(1 + np.exp(-x))

    def _tanh(self, x: T, derivative=False) -> T:
        tan = 2/(1 + np.exp(-2*x)) - 1
        if derivative:
            return (1-tan**2)
        return tan

    def _relu(self, x: T, derivative=False) -> T:
        if derivative:
            return (x > 0).astype(int)
        return np.max(0, x)

    def _softmax(self, x: T, derivative=False) -> T:
        if derivative:
            return x * (1 - x)
        # np.exp(x)/np.sum(np.exp(x))
        exponentials = np.exp(x)
        total = np.sum(exponentials)
        return exponentials/total

    def _identity(self, x: T, derivative=False) -> T:
        if derivative:
            return np.ones_like(x)
        return x



import typing
import numpy as np

T_Feature_Array = np.ndarray
T_Target_Array = np.ndarray
T_Output_Array = T_Target_Array

T_Data_Sample = typing.Tuple[T_Feature_Array, T_Target_Array]
T_Dataset = typing.List[T_Data_Sample]



import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
import numpy as np
from typing import List
from .Types import T_Feature_Array, T_Target_Array, T_Data_Sample, T_Dataset


# Dataset class
class Dataset():
    def __init__(self, I, O):
        """
            Dataset Constructor

            Parameters
            ----------
            I: int
                No of inputs
            O: int
                No of outputs

        """
        self.I = I
        self.O = O
        self.size = 0
        self.dataset: T_Dataset = list()

    def make_dataset(self, input_file, target_file):
        """
            Creates a dataset

            Parameters
            ----------
            input_file: str
                csv file containing the features/inputs.
            target_file: str
                csv file containing the targets.

            Returns
            -------
            Doesn't return anything.
        """

        input_handler = self.open_file(input_file)
        target_handler = self.open_file(target_file)

        if not input_handler or not target_handler:
            print("Unable to create Dataset")
            return
        input_lines = input_handler.readlines()
        target_lines = target_handler.readlines()

        for inp, tar in zip(input_lines, target_lines):
            input_array = list(map(float, inp.split(',')))
            target_array = list(map(float, tar.split(',')))

            features: T_Feature_Array = np.reshape(input_array, (self.I, 1))
            targets: T_Target_Array = np.reshape(target_array, (self.O, 1))
            sample: T_Data_Sample = (features, targets)
            self.dataset.append(sample)
            self.size += 1

    def modify_lists(self, input_array, target_array):
        sample = list()
        sample.append(np.reshape(input_array, (self.I, 1)))
        sample.append(np.reshape(target_array, (self.O, 1)))
        self.dataset.append(sample)
        self.size += 1

    def get_raw_data(self):
        """
            Returns the dataset which was made earlier in make_dataset method

            Parameters
            ----------
            Doesn't accept anything

            Returns
            Tuple[T_Dataset, int]
                Dataset and its size
        """
        return self.dataset, self.size

    def display(self):
        """
            Prints the dataset
        """
        for i in range(self.size):
            sample = self.dataset[i]
            print("Data Sample:", i+1)
            print("\tInput: ", sample[0])
            print("\tTarget: ", sample[1])

    @staticmethod
    def open_file(filename):
        """
            Just a helper function to open a given file and handle errors if any.

            Parameters
            ----------
            filename: str
                Filename to be opened

            Returns
            -------
            fhand
                A filehandler corresponding to the given file.
        """
        try:
            fhand = open(filename)
        except Exception as e:
            print("[!!] Exception Occurred while Opening", filename, e)
            return None
        return fhand

    def get_min_max_values(self, array: T_Dataset):
        """
            Computes the min and max of each feature, and each target label of the entire dataset.

            Parameters
            ----------
            array : List[List[np.array]]
                List of datasamples
                datasample = [
                    column vector of features,
                    column vector of of targets
                ]

            Returns
            -------
            Tuple[List[float]]
                min and max values of features and targets
                (min_max_of_features, min_max_of_targets)
                min_max_of_features = List[[min_of_ith_feature, max_of_ith_feature]]
                min_max_of_targets = List[[min_of_ith_target, max_of_ith_target]]
        """

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features = list()
        # print(min_max_of_features)
        for i in range(no_of_features):
            ith_feature_values = [sample[0][i][0] for sample in array]
            # print(ith_feature_values)
            min_of_ith_feature = min(ith_feature_values)
            max_of_ith_feature = max(ith_feature_values)
            min_max_of_features.append(
                [min_of_ith_feature, max_of_ith_feature])

        # print(min_max_of_features)

        min_max_of_targets = list()
        # print(min_max_of_targets)
        for i in range(no_of_targets):
            ith_target_values = [sample[1][i][0] for sample in array]
            # print(ith_target_values)
            min_of_ith_target = min(ith_target_values)
            max_of_ith_target = max(ith_target_values)
            min_max_of_targets.append([min_of_ith_target, max_of_ith_target])

        # print(min_max_of_targets)

        return min_max_of_features, min_max_of_targets

    def scale_data(self, array: T_Dataset, size):
        """
            Scales the data using min-max scaling method.

            parameters
            ----------
            array: T_Dataset
                Dataset to be scaled.

            size: int
                Size of the given dataset.

            Returns
            -------
            array: T_Dataset
                Scaled dataset.
        """

        if size == 0:
            return

        # print(array)

        no_of_features = array[0][0].shape[0]
        no_of_targets = array[0][1].shape[0]

        min_max_of_features, min_max_of_targets = self.get_min_max_values(
            array)

        for i in range(size):
            sample = array[i]
            features = sample[0]
            targets = sample[1]

            for j in range(no_of_features):
                min_of_jth_feature = min_max_of_features[j][0]
                max_of_jth_feature = min_max_of_features[j][1]
                features[j][0] = (features[j][0] - min_of_jth_feature) / \
                    (max_of_jth_feature - min_of_jth_feature)

            for j in range(no_of_targets):
                min_of_jth_target = min_max_of_targets[j][0]
                max_of_jth_target = min_max_of_targets[j][1]
                targets[j][0] = (targets[j][0] - min_of_jth_target) / \
                    (max_of_jth_target - min_of_jth_target)

        return array
